// Demonstration of a mutex.
//
// Author: Sol Boucher <sboucher@cmu.edu>

#include <stdio.h>
#include <unistd.h>

#include "pthread.h"

#define NR_THREADS 2
#define DELAY_SECS 2

static void *thread(void *ptr) {
	pthread_mutex_t *lock = (pthread_mutex_t *) ptr;

	if(lock)
		pthread_mutex_lock(lock);
	printf("Thread %lu: BEGIN critical section\n", pthread_self());
	sleep(DELAY_SECS);
	printf("Thread %lu: END critical section\n", pthread_self());
	sleep(DELAY_SECS);
	if(lock)
		pthread_mutex_unlock(lock);

	return NULL;
}

static inline void spawn_threads(pthread_mutex_t *lock) {
	pthread_t tid[NR_THREADS];

	for(int td = 0; td < NR_THREADS; ++td)
		pthread_create(tid + td, NULL, thread, lock);
	for(int td = 0; td < NR_THREADS; ++td)
		pthread_join(tid[td], NULL);
}

int main(void) {
	puts("Trying WITHOUT lock:");
	spawn_threads(NULL);
	putchar('\n');

	pthread_mutex_t lock;
	pthread_mutex_init(&lock, NULL);

	puts("Trying WITH lock:");
	spawn_threads(&lock);

	pthread_mutex_destroy(&lock);
	return 0;
}
