// Demonstration of a reader-writer lock. Note how readers are able to
// "cut in line": there's no queue in the rwlock, so requests are serviced in an
// arbitrary order!
//
// Author: Sol Boucher <sboucher@cmu.edu>

#include <stdio.h>
#include <unistd.h>

#include "pthread.h"

#define NR_READERS 2
#define NR_WRITERS 2
#define DELAY_SECS 2

static void *reader(void *ptr) {
	pthread_rwlock_t *lock = (pthread_rwlock_t *) ptr;

	if(lock)
		pthread_rwlock_rdlock(lock);
	printf("Reader %lu: BEGIN critical section\n", pthread_self());
	sleep(DELAY_SECS);
	printf("Reader %lu: END critical section\n", pthread_self());
	sleep(DELAY_SECS);
	if(lock)
		pthread_rwlock_unlock(lock);

	return NULL;
}

static void *writer(void *ptr) {
	pthread_rwlock_t *lock = (pthread_rwlock_t *) ptr;

	if(lock)
		pthread_rwlock_wrlock(lock);
	printf("Writer %lu: BEGIN critical section\n", pthread_self());
	sleep(DELAY_SECS);
	printf("Writer %lu: END critical section\n", pthread_self());
	sleep(DELAY_SECS);
	if(lock)
		pthread_rwlock_unlock(lock);

	return NULL;
}

static inline void spawn_threads(pthread_rwlock_t *lock) {
	const int NR_THREADS = 2 * NR_READERS + NR_WRITERS;
	pthread_t tid[NR_THREADS];

	for(int td = 0; td < NR_READERS; ++td)
		pthread_create(tid + td, NULL, reader, lock);
	for(int td = NR_READERS; td < NR_READERS + NR_WRITERS; ++td)
		pthread_create(tid + td, NULL, writer, lock);
	for(int td = NR_READERS + NR_WRITERS; td < NR_THREADS; ++td)
		pthread_create(tid + td, NULL, reader, lock);
	for(int td = 0; td < NR_THREADS; ++td)
		pthread_join(tid[td], NULL);
}

int main(void) {
	puts("Trying WITHOUT lock:");
	spawn_threads(NULL);
	putchar('\n');

	pthread_rwlock_t lock;
	pthread_rwlock_init(&lock, NULL);

	puts("Trying WITH lock:");
	spawn_threads(&lock);

	pthread_rwlock_destroy(&lock);
	return 0;
}
