// Demonstration of a semaphore.
//
// Author: Sol Boucher <sboucher@cmu.edu>

#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

#include "semaphore.h"

#define INIT_VALUE 2
#define NR_THREADS 3
#define DELAY_SECS 2

static void *thread(void *ptr) {
	sem_t *lock = (sem_t *) ptr;

	if(lock)
		sem_wait(lock);
	printf("Thread %lu: BEGIN critical section\n", pthread_self());
	sleep(DELAY_SECS);
	printf("Thread %lu: END critical section\n", pthread_self());
	sleep(DELAY_SECS);
	if(lock)
		sem_post(lock);

	return NULL;
}

static inline void spawn_threads(sem_t *lock) {
	pthread_t tid[NR_THREADS];

	for(int td = 0; td < NR_THREADS; ++td)
		pthread_create(tid + td, NULL, thread, lock);
	for(int td = 0; td < NR_THREADS; ++td)
		pthread_join(tid[td], NULL);
}

int main(void) {
	puts("Trying WITHOUT lock:");
	spawn_threads(NULL);
	putchar('\n');

	sem_t lock;
	sem_init(&lock, 0, INIT_VALUE);

	puts("Trying WITH lock:");
	spawn_threads(&lock);

	sem_destroy(&lock);
	return 0;
}
